########################################################################
########################################################################
#### Trace formula routines.
########################################################################
########################################################################

### Compute the coefficients of the characteristic power series of U_p acting 
### in tame level N on the rth component of the p-adic weight space
# Output:
#	A list [a_0,a_1,...,a_jmax] where each a_i is a one-variable power series over Zp, seen
#	then as rigid functions on the rth component of weight space, such that
#	det(1 - XUp) = a0 + a1*X + a2*X^2 + ... is the characteristic power series of Up acting in tame
#	level N on weights in the rth component of weight space
# Input: 
# 	N: Tame level
#	p: prime that is co-prime to N
#	j_max: Integer >= 0
#	r: An integer 0 <= r <= p-1 ("component of weight space")
#	cusp: True or False, whether or not to just compute with CuspForms or whether
#			to include all the Eisenstein series
#	gamma: A topological generator for 1+2pZ_p (default = -1 but that will become 1+p, or 5, depending 
#			on if p != 2 or not)
#	M1: p-adic accuracy (default to 200)
#	M2: T-adic accuracy (in the weight variable, default to 100)
#	verbose: true/false, whether or not to output what index j is currently being computed (default = false)

def Up_charseries_coeffs(N,p,j_max,r,cusp,gamma=-1,M1=200,M2=100,verbose=false):
	# first get tr(Up^j) for j = 0,1,...,jmax
	t_list = traceUp_list(N,p,j_max,r,cusp,gamma,M1,M2,verbose)
	# now use Newton's identities to turn sums of power of eigenvalues of Up into
	# sums of eigenvalues for wedge powers of Up 
	b_list = powertosymmetric(t_list,M2)
	# now make sure the signs are alternating to get the true characteristic power series
	a_list = [(-1)^i*b_list[i] for i in range(len(b_list))]
	return a_list
	
	

### Helper for the above routine
### It is Newton's identity for symmetric functions or something like that
# Output:
#	If we write ti = lambda_1^i + .... + lambda_n^i + .... then this will output
#   the list ai = sum of ways to write lambda_j1*lambda_j2*...*lambda_ji (i times)
# Input:
#	t_list: A list t_list = [t0,t1,...] of one variable (T) power series in t
#	max: an integer T >= 0

def powertosymmetric(t_list,max):
	M = len(t_list)
	s = [1]
	for n in range(1,M):
		s.append(-1/n * sum([(-1)^(n-r) * trunc(s[r]*t_list[n-r],max) for r in range(n)]))
	return s

# Output:
#	The list of functions tr(U_p^j), j=0,1,...,j_max seen as rigid analytic functions on the rth component
#	of the p-adic weight space. Thus it is a list of one-variable power series over Z_p.
# Input: 
# 	N: Tame level
#	p: prime that is co-prime to N
#	j_max: Integer >= 0
#	r: An integer 0 <= r <= p-1 ("component of weight space")
#	cusp: True or False, whether or not to just compute with CuspForms or whether
#			to include all the Eisenstein series
#	gamma: A topological generator for 1+2pZ_p (default = -1 but that will become 1+p, or 5, depending 
#			on if p != 2 or not)
#	M1: p-adic accuracy
#	M2: T-adic accuracy (in the weight variable)
#	verbose: true/false, whether or not to output what index j is currently being computed (default = false)

def traceUp_list(N,p,j_max,r,cusp,gamma=-1,M1=200,M2=100,verbose=False):
	t_list = [0]
	if gamma == -1:
		if p == 2:
			gamma = 5
		else:
			gamma = 1+p
	if cusp:
		cusp_type = 'cuspidal'
	else:
		cusp_type = 'modular'
	for i in range(1,j_max):
		if verbose:
			print "\t\t Computing trace(U_" + str(p) + "^" + str(i) + ") in " + str(cusp_type) + " level N = " + str(N) + " on component r = " + str(r)
		t_list.append(traceUp(N,p,i,r,cusp,gamma,M1,M2))
	return t_list


### Compute the trace of a single p-adic Hecke operator in tame level
### away from p.
# Output:
#	The function tr(U_p^n) seen as a rigid analytic function on the rth component
#	of the p-adic weight space. Thus it is a one-variable power series over Z_p.
# Input: 
# 	N: Tame level
#	p: prime that is co-prime to N
#	n: Integer >= 0
#	r: An integer 0 <= r <= p-1 ("component of weight space")
#	cusp: True or False, whether or not to just compute with CuspForms or whether
#			to include all the Eisenstein series
#	gamma: A topological generator for 1+2pZ_p (default = -1 but that will become 1+p, or 5, depending 
#			on if p != 2 or not)
#	M1: p-adic accuracy
#	M2: T-adic accuracy (in the weight variable)

def traceUp(N,p,n,r,cusp,gamma=-1,M1=200,M2=100):
	if gamma == -1:
		if p == 2:
			gamma = 5
		else:
			gamma = 1+p
	t = 1
	ans = 0
	if cusp:
		ans = const_term(p,n,N)
	while 4*p^n-t^2 > 0:
		if t % p != 0:
			D = t^2 - 4*p^n
			e = get_t(t,p^n)
			class_num_cont = sum([norm_class_num(D/f^2)*c(p^n,N,t,f) for f in divisors(e)])
			R.<x> = PolynomialRing(pAdicField(p,M1))
			f = x^2 - D
			sD = f.roots()[0][0]
			if (sD-3).valuation > 1:
				sD = f.roots()[1][0]
			rho = (-t + sD)/2
			if rho.valuation() > 0:
				rho = (-t - sD)/2
			rhob = p^n/rho
			#ans = ans - (rho * (rho-rhob))^(-1) * QQ(pari(-D).qfbhclassno()) * kth_power_ps(rho,p,r,gamma,M1,M2)
			ans = ans - (rho * (rho-rhob))^(-1) * class_num_cont * kth_power_ps(rho,p,r,gamma,M1,M2)
		t = t + 1
	return ans




########################################################################
########################################################################
#### Helpers for turning Iwasawa ring elements into power series
#### over Z_p
########################################################################
########################################################################
    
# Returns log_p(x) computed with M terms

def logp(x,p,M):
    R = pAdicRing(p,M)
    y = x / R.teichmuller(x)
    return sum([(y-1)^j*(-1)^(j+1)/j for j in range(1,M+1)])
    

# Takes a power series f in Z_p[[T]] and returns the truncation
# of f to degree <= M

def trunc(f,M):
    T = f.parent().gens()[0]
    return(sum([f.coefficients()[r] * T^r for r in range(min(M,len(f.coefficients())))]))
    

# For a in 1+pZ_p, returns the power series representation of [a] in Z_p[[1+pZ_p]] = Z_p[[T]] 
# using a top. generator gamma of 1+pZ_p (or 1 + 4Z_2 if p = 2)
# Warning: Accuracy not well handled.  
# M1 represents number of terms to compute logs (p-adic accuracy);
# M2 is degree in T to compute out to

def grplike_to_ps(a,p,gamma,M1,M2):
    logpgamma = logp(gamma,p,M1)
    R.<T> = PolynomialRing(pAdicField(p,M1))
    return sum([binomial(ZZ(logp(a,p,M1)/logpgamma),n) * T^n for n in range(M2)])
    

# Given a in Z_p^x, returns the powers representation of the map k --> a^k as k varies over the r-th component of weight space
# As above: gamma is a top. generator of 1+pZ_p, M1 is a p-adic accuracy and M2 is a T-adic accuracy

def kth_power_ps(a,p,r,gamma,M1,M2):
    R = pAdicRing(p,M1)
    if p > 2:
        return R.teichmuller(a)^r * grplike_to_ps(a/R.teichmuller(a),p,gamma,M1,M2)
    else:
        if r == 0:
            if a % 4 == -1:
                b = -a
            else:
                b = a
            return grplike_to_ps(b,p,gamma,M1,M2)
        else:
            print "Oh oh"

########################################################################
########################################################################
#### Helpers for computing some class number returns
########################################################################
########################################################################

## returns the normalized class number of the *imaginary* quadratic order
## of discriminant D
## Must have D<0 and D = 0,1 mod 4. then write D = f^2d_K where d_K
## is a fundamental discriminant. The class number is normalized by
## \hat h(D) = h(D)/w where w is half the size of the unit group ...
def norm_class_num(D):
    # first step is to write D = f^2 d with d a fundamental discriminant
    if (D % 4) not in [0,1]:
        return 0
    d = D.squarefree_part()
    if d % 4 == 1:
        f = ZZ(sqrt(D/d))
    if d % 4 != 1:
        f = ZZ(sqrt(D/(4*d)))
    d = D/f^2
    # now D = f^2 d or f^2 4d
    if f == 1:
        if d == -3:
            return QuadraticField(d).class_number()/3
        elif d == -4:
            return QuadraticField(d).class_number()/2
        else:
            return QuadraticField(d).class_number()
    else:
        return f*norm_class_num(d)*prod([1-kronecker(d,p)/p for p in f.prime_divisors()])

## returns the Hurtwiz class number H(n)
## by H(0) = -1/2, H(n) = 0 if n < 0 and H(n) otherwise counts the number
## of positive definite binary quadratic forms with some multiplicity
def hurwitz_class_no(n):
    if n < 0:
        return 0
    elif n == 0:
        return -1/12
    else:
        return sum([norm_class_num(-n/f^2) for f in n.divisors() if n % f^2 == 0])

## given an s and n, let Delta = s^2 - 4*n. If Delta < 0 then it returns
## the unique non-negative integer t such that Delta = t^2 d_K where d_K is
## a discriminant of an imaginary quadratic field        
def get_t(s,n):
    delta = s^2 - 4*n
    if delta == 0:
        return 0
    m = delta.squarefree_part()
    if m < 0 and m % 4 != 1:
        t = sqrt(ZZ(delta/(4*m)))
    else:
        t = sqrt(ZZ(delta/m))
    return t

########################################################################
########################################################################
#### Routines to compute the explicit trace formula 
#### as it appears in Koike (1975) and Hijikata (1974)
#### all notation is taken from Hijikata p. 57-58
#### In general a tame level N is given and the index m, co-prime to N
#### of a Hecke operator acting on level Gamma0(N) is given as well.
#### In practice, m will be a power of p co-prime to N.
########################################################################
########################################################################    
    
    
## computes \twid A
## input:
##		m = index of Hecke operator
## 		N = (tame) level
##		s = as in Hijikata
##		f = a divisor of the square part of s^2 - 4*m
##		ell = a prime divisor of N

def twid_A(m,N,s,f,ell,verbose=false):
    nu = N.valuation(ell)
    rho = f.valuation(ell)
    R.<X> = PolynomialRing(ZZ)
    Phi = R(X^2 - s*X + m)
    modulus = ell^(nu+2*rho)
    res = [x % modulus for x in range(modulus) if Phi(x) % modulus == 0 and (2*x-s) % ell^rho == 0]
    if verbose:
        print "twid_A = ", res
    return res

## computes \twid B
## input:
##		m = index of Hecke operator
## 		N = (tame) level
##		s = as in Hijikata
##		f = a divisor of the square part of s^2 - 4*m
##		ell = a prime divisor of N
 
def twid_B(m,N,s,f,ell,verbose=false):
    nu = N.valuation(ell)
    rho = f.valuation(ell)
    R.<X> = PolynomialRing(ZZ)
    Phi = R(X^2 - s*X + m)
    modulus = ell^(nu+2*rho+1)
    res = [x % modulus for x in range(modulus) if ( Phi(x) % modulus == 0) and ((2*x-s) % ell^rho == 0) ] 
    if verbose:
        print "twid_B = ", res   
    return res

## computes A = \twid A (mod ell^(nu+rho)) where
## nu = v_ell(N) and rho = v_ell(f)
## input: same as \twid A
    
def A(m,N,s,f,ell,verbose=false):
    nu = N.valuation(ell)
    rho = f.valuation(ell)
    return list(set([x % ell^(nu+rho) for x in range(ell^(nu+rho)) if x % ell^(nu+2*rho) in twid_A(m,N,s,f,ell,verbose)]))

## computes B = \twid B (mod ell^(nu+rho)) where
## nu = v_ell(N) and rho = v_ell(f)
## input: same as \twid B

def B(m,N,s,f,ell,verbose=false):
    nu = N.valuation(ell)
    rho = f.valuation(ell)
    return list(set([x % ell^(nu+rho) for x in range(ell^(nu+rho)) if x % ell^(nu+2*rho) in twid_B(m,N,s,f,ell,verbose)]))

## computes A_eta = {x in A : x = eta mod ell^rho} (rho = v_ell(f))
## input: same as \twid A
## 			eta an element of (Z/NZ)^x

def A_eta(m,N,s,f,ell,eta,verbose=false):
    nu = N.valuation(ell)
    return [x for x in A(m,N,s,f,ell,verbose) if (x - eta) % ell^nu == 0]

## computes B_eta = {s-x : x in B and s-x = eta mod ell^rho} (rho = v_ell(f))
## input: same as \twid A
## 			eta an element of (Z/NZ)^x

def B_eta(m,N,s,f,ell,eta,verbose=false):
    nu = N.valuation(ell)
    return [s-x for x in B(m,N,s,f,ell,verbose) if (s-x - eta) % ell^nu == 0]

## computes c_eta which counts A_eta and B_eta,
## depending on whether or not (s^2-4*m)/f^2 is an ell-adic unit or not
def c_eta(m,N,s,f,ell,eta,verbose=false):
    Delta = s^2 - 4*m
    check = ZZ(Delta/f^2)
    if check.valuation(ell) != 0:
        res = len(A_eta(m,N,s,f,ell,eta,verbose)) + len(B_eta(m,N,s,f,ell,eta,verbose))
    else:
        res = len(A_eta(m,N,s,f,ell,eta,verbose))
    return res

## computes c = sum_eta (prod_{ell | N} c_eta)
##     
def c(m,N,s,f,verbose=false):
	#L = CyclotomicField(euler_phi(N))
	#chi_list = [chi for chi in DirichletGroup(N) if chi(-1) == 1]
	inv_etas = [eta for eta in range(N) if gcd(eta,N) == 1]
	result = 0
	for eta in inv_etas:
		cont = prod([c_eta(m,N,s,f,ell,eta,verbose) for ell in prime_divisors(N)])
		#result += sum([chi(eta)*L(cont) for chi in chi_list])
		result += cont
	return QQ(result)

## returns the a priori "constant term" in the trace formula
## for tr(U_p^j) acting in tame level N
    
def const_term(p,j,N,verbose=false):
	if N == 1:
		return - 1
	else:
		return (-1/(p^j-1))*sum([euler_phi((p^j-1)/f)*c(p^j,N,p^j+1,f,verbose) for f in divisors(p^j-1)])