########################################################################
########################################################################
#### Routines for how Bergdall--Pollack calculated the data presented
#### in Section 4 of the preprint "Arithmetic properties of Fredholm series
#### for p-adic modular forms"
########################################################################
########################################################################


# We require the file koike.sage that contains
# the routines to compute the characteristic power series

load('koike.sage')


# Output:
#	The mu,lambda and location of zeroes information for the first j_max-many 
#	coefficients of the characteristic series of Up acting in tame level N on the rth
#	component of the p-adic weight space.
# Input: 
# 	N: Tame level
#	p: prime that is co-prime to N
#	j_max: Integer >= 0
#	r: An integer 0 <= r <= p-1 ("component of weight space")
#	cusp: True or False, whether or not to just compute with CuspForms or whether
#			to include all the Eisenstein series
#	gamma: A topological generator for 1+2pZ_p (default = -1 but that will become 1+p, or 5, depending 
#			on if p != 2 or not)
#	M1: p-adic accuracy (default to 200)
#	M2: T-adic accuracy (in the weight variable, default to 100)
#	verbose: true/false, whether or not to output what index j is currently being computed (default = false)
#	write: true/false, whether or not to write a file
#	filename: a specified file name. If you do not specify the filename then we will provide it for you in an obvious form
#		For example: cusp_p=2_N=9_jmax=10_r=0.txt

	
def mu_lambda_slopes_data(N,p,j_max,r,cusp=0,gamma=-1,M1=200,M2=100,verbose=False,write=False,filename=''):
	if cusp:
		cusp_type = 'cusp'
	else:
		cusp_type = 'mod'
	if write:
		if filename == '':
			filename = str(cusp_type) + "_p=" + str(p) + "_N=" + str(N) + "_jmax=" + str(j_max) + "_r=" + str(r) + ".txt"
		F = open(filename,'w')
	
	str1 = str(cusp_type) + "\n p="+str(p)+", N="+str(N)+", j_max="+str(j_max)+", r="+str(r)+"\n\n"
	
	if verbose:
		print str1 + "Getting the trace of powers of U_" + str(p) + " at tame + " + str(cusp_type) + " level N = " + str(N) + "\n"
	if write:
		F.write(str1)
		
	a_list = Up_charseries_coeffs(N,p,j_max,r,cusp,gamma,M1,M2,verbose)
	str5 = "\nCoefficient data:\n"
	if verbose:
		print str5
	if write:
		F.write(str5)

	a_data = mu_lambda_slopes_list(a_list,M1)
	str6 = print_pretty_data(a_data)
	if verbose:
		print str6
	if write:
		F.write(str6)
	
	if write:
		F.close()


# Output:
#	A list of triples (mu,lambda,[slopes]) where the ith entry in the list 
#	is mu = mu(pi), lambda = lambda(pi) and the list of slopes
#	are the locations (in terms of p-adic valuation) of the roots of pi in
#	the open unit disc.
# Input:
#	p_series: A list of power series [p0,p1,...]	
#	M: The T-adic accuracy that the power series pi \in Zp[[T]] were computed to
#	pretty_output: true/false. If true then will also output to the screen the data
#		in a readable fashion. This uses the helper routine below.
# Warning: if there is not enough p-adic accuracy in the computation of the pi
# then the program may not compute the location of the actual zeroes 

def mu_lambda_slopes_list(p_series,M=100):
	res = []
	for j in range(len(p_series)):
		try:
			slopes = p_series[j].newton_slopes()
		except:
			slopes = []
		res.append([muinv(p_series[j],M), lambdainv(p_series[j],M), slopes])
	return res
	
# Helper for the above routine(s).
# Input:
#	data_list: [ [mu,lambda,[slopes]]_j : j in ...]
# Output:
#	s: A string the when printed to the screen or a file will output
#		the mu,lambda,... information in a readable fashion

def print_pretty_data(data_list,trunc=True):
	L = len(data_list)
	s = ''
	for j in range(L):
		mu = data_list[j][0]
		lamb = data_list[j][1]
		if trunc:
			try:
				slopes = data_list[j][2][0:lamb]
			except:
				slopes = data_list[j][2]
		else:
			slopes = data_list[j][2]
		s += str(j) + ":\n"
		s += "\t mu = " + str(mu) + "\n"
		s += "\t lm = " + str(lamb) + "\n"
		s += "\t slopes:" + str(slopes) + "\n"
	return s
	
	


########################################################################
########################################################################
#### Helper routines to compute mu and lambda-invariants
#### of a one variable power series over Zp
########################################################################
########################################################################

### Returns the mu-invariant of a power series over Zp
# Output: 
#	The highest power mu of p such that p^mu divides f. If f = u0 + u1*T + ... then
#	with ui in Zp then mu = min(val(ui)). It will return +Infinity if f = 0.
# Input:
#	f: A one variable power series over Zp
#	M: Only examines first M coefficients (useful if you aren't sure how many coefficients
#		of f are p-adically accurate)

def muinv(f,M):
	try:
		coeffs = f.coefficients()
		return min([coeffs[r].valuation() for r in range(min(M,len(coeffs)))])
	except:
		return +Infinity
### Returns the lambda-invariant of a power series over Zp
# Output:
#	The integer lambda >= 0 that represents the number of zeroes of f in the open
#	unit disc |T| < 1. If f = u0 + u1*T + ... then lambda is the first index i such that
#	val(ui) = mu(f) where mu(f) is the mu-invariant of f.
# Input:
#	f: A one variable power series over Zp
#	M: Only examines first M coefficients (useful if you aren't sure how many coefficients
#		of f are p-adically accurate)

def lambdainv(f,M):
	mu = muinv(f,M)
	try:
		coeffs = f.coefficients()
		v = [coeffs[r].valuation()-mu for r in range(min(M,len(coeffs)))]
		return v.index(0)
	except:
		return 0

